package require Tkogl
package require OpenGL
package require OpenGLU

set debug [open "c:\\debug.txt" w]

# Create OpenGL window
pack [OGLwin .gl -refresh "draw .gl" -width 300 -height 300 ] -expand yes -fill both



set pickdot {}
set dots {}

# Initialization of viewing parameters
.gl makeCurrent

glMatrixMode GL_PROJECTION
glLoadIdentity
glOrtho -5 5 -5 5 -5 5
glMatrixMode GL_MODELVIEW
glLoadIdentity
glEnable GL_MAP1_VERTEX_3
glMapGrid1f 100 0 1

# Initial control point coordinates
set cpoints [array_double 12]

set points(0) {-4 -4 0}   
set points(1) {-2 4 0}
set points(2) {2 -4 0}
set points(3) {4 4 0}

for {set i 0} {$i < 4} {incr i} {
	set_double $cpoints [expr $i * 3] [lindex $points($i) 0]
	set_double $cpoints [expr $i * 3 + 1] [lindex $points($i) 1]
	set_double $cpoints [expr $i * 3 + 2] [lindex $points($i) 2]
}

set curve [glGenLists 1]
set dots [glGenLists 1]

# Code to render curve and control points
proc computecurve {} {
	global curve cpoints dots

	.gl makeCurrent
    
	catch {glDeleteLists $curve 1}
	catch {glDeleteLists $dots 1}

	glNewList $curve GL_COMPILE
	glColor3f 0 1 1
	glMap1d GL_MAP1_VERTEX_3 0 1 3 4 $cpoints
	glEvalMesh1 GL_LINE 0 100
	glEndList	

	glNewList $dots GL_COMPILE
	glColor3f 1 0 0
	glInitNames 
	glPushName -1 
	glPointSize 4

	for {set i 0} {$i < 4} {incr i} {
		glLoadName $i 
		glBegin GL_POINTS
			glVertex3d [get_double $cpoints [expr $i *3]] [get_double $cpoints [expr ($i * 3) + 1]] [get_double $cpoints [expr ($i * 3) + 2]]
		glEnd
	}
    
	glEndList
}

# Code to figure out which control point was picked
proc pick_point { x y } {
	global cpoints dots pickdot debug

	puts $debug "pick_point $x $y"

	set Buffer [array_int 100]
	set viewport [array_int 4]

	.gl makeCurrent

	glSelectBuffer 100 $Buffer
	glGetIntegerv GL_VIEWPORT $viewport

	puts $debug "[get_int $Buffer 0] [get_int $Buffer 1] [get_int $Buffer 2] [get_int $Buffer 3]"
	
	glMatrixMode GL_PROJECTION
	glPushMatrix
	glLoadIdentity

	glRenderMode GL_SELECT	
	
	gluPickMatrix $x [expr [get_int $viewport 3] - $y] 4 4 $viewport
	glOrtho -5 5 -5 5 -5 5
	glMatrixMode GL_MODELVIEW
	glCallList $dots
	glMatrixMode GL_PROJECTION
	glPopMatrix
	glMatrixMode GL_MODELVIEW

	glFlush

	set pickdot {}
	set nHits [glRenderMode GL_RENDER]
	puts $debug "nHits = $nHits"
	
	for {set i 0} {$i < $nHits} {incr i} {
		puts $debug "Buffer($i) = [get_int $Buffer $i]"
		set vals [get_int $Buffer $i]
		incr i
		puts $debug "Buffer($i) = [get_int $Buffer $i]"
		incr i
		puts $debug "Buffer($i) = [get_int $Buffer $i]"
		incr i
		puts $debug "Buffer($i) = [get_int $Buffer $i]"
		for {set j 0} {$j < $vals} {incr j} {
			incr i
			puts $debug "Buffer($i) = [get_int $Buffer $i]"
		}
	}

	if {$nHits == 1} { set pickdot [get_int $Buffer 3] }

}

# Code to drag the picked control point
proc move_point { x y } {
	global cpoints pickdot debug
	
	.gl makeCurrent

	if { $pickdot != {} } {
		set xpos [expr $pickdot * 3]
		set ypos [expr ($pickdot * 3) + 1]
		set zpos [expr ($pickdot * 3) + 2]
		set newPos [.gl unproject $x $y 0]

		set_double $cpoints $xpos [lindex $newPos 0]
		set_double $cpoints $ypos [lindex $newPos 1]
		set_double $cpoints $zpos [lindex $newPos 2]
		computecurve
		.gl redraw
	}
}

proc draw {path} {
	global cpoints curve dots

	glClear GL_COLOR_BUFFER_BIT
	glColor3f 0 1 1

	glCallList $curve
	glCallList $dots

	$path swapBuffer
}

computecurve

bind .gl <Button-1> { pick_point %x %y }
bind .gl <B1-Motion> { move_point %x %y }
