package require Tkogl
package require OpenGL



OGLwin .gl -refresh "refresh .gl" -height 400 -width 400
pack .gl

OGLwin .gl1 -refresh "refresh .gl1" -height 400 -width 400
pack .gl1

set xrot 0.0
set yrot 0.0
set zrot 0.0 

set xrot1 0.0
set yrot1 0.0
set xrot2 0.0
set yrot2 0.0

set specular1 [newfv4 1 1 1 1]
set specular2 [newfv4 0 1 1 1]

proc colour {val} {
	global ambientlight red green blue

	set_float $ambientlight 0 $red
	set_float $ambientlight 1 $green
	set_float $ambientlight 2 $blue

	.gl redraw
	.gl1 redraw
}

proc initWindow {win speclight} {
	global ambientlight shiny red green blue

	set ambientlight [newfv4 0.5 0.5 0.5 1.0]
	set red 0.5
	set green 0.5
	set blue 0.5
	set shiny 20

	set nx 4
	set ny 4
	set controlPts [array_double [expr 4 * 4 * 3]]

	set c(1) {-1.5 -1.5 4}   
	set c(2) {-0.5 -1.5 2.0}
	set c(3) {0.5 -1.5 -1.0}
	set c(4) {1.5 -1.5 2.0}
	set c(5) {-1.5 -0.5 1.0}
	set c(6) {-0.5 -0.5 3}
	set c(7) {0.5 -0.5 0}
	set c(8) {1.5 -0.5 3}
	set c(9) {-1.5 0.5 4.0} 
	set c(10) {-0.5 0.5 0}
	set c(11) {0.5 0.5 3}
	set c(12) {1.5 0.5 4}
	set c(13) {-1.5 1.5 -2.0} 
	set c(14) {-0.5 1.5 -2.0}
	set c(15) {0.5 1.5 0}
	set c(16) {1.5 1.5 -1}

	for {set i 1} {$i < 17} {incr i} {
		set_double $controlPts [expr ($i-1)* 3] [lindex $c($i) 0]
		set_double $controlPts [expr ($i-1)* 3 + 1] [lindex $c($i) 1]
		set_double $controlPts [expr ($i-1)* 3 + 2] [lindex $c($i) 2]
	}

	set vstride [expr $nx*3]

	$win makeCurrent
	glMatrixMode GL_PROJECTION
	glLoadIdentity 
	glOrtho -4 4 -4 4 -4 4 
	glMap2d GL_MAP2_VERTEX_3 0 1 3 $nx 0 1 $vstride $ny $controlPts
	glEnable GL_MAP2_VERTEX_3 
	glEnable GL_AUTO_NORMAL
	glMaterialfv GL_FRONT_AND_BACK GL_AMBIENT_AND_DIFFUSE $ambientlight 
	glMaterialfv GL_FRONT_AND_BACK GL_SPECULAR $speclight 
	glMateriali GL_FRONT_AND_BACK GL_SHININESS $shiny
	glShadeModel GL_SMOOTH
	glFrontFace GL_CW
	glLightModeli GL_LIGHT_MODEL_TWO_SIDE 1 
	glEnable GL_LIGHT0
	glEnable GL_LIGHTING
	glEnable GL_DEPTH_TEST
}

initWindow .gl $specular1
initWindow .gl1 $specular2

proc dome {val} {
# this is a dummy function that is called everytime the scale is changed
# so that we can strip off the argument that is passed
	dome
}

proc refresh {win} {
	global xrot1 yrot1 xrot2 yrot2 zrot ambientlight shiny

	if {$win == ".gl"} {
		set xrot $xrot1
		set yrot $yrot1
	} else {
		set xrot $xrot2
		set yrot $yrot2
	}

	$win makeCurrent

	glMaterialfv GL_FRONT_AND_BACK GL_AMBIENT_AND_DIFFUSE $ambientlight 
	glMateriali GL_FRONT_AND_BACK GL_SHININESS $shiny
	glMatrixMode GL_MODELVIEW
	glLoadIdentity
	glRotatef $xrot 1.0 0.0 0.0
	glRotatef $yrot 0.0 1.0 0.0
	glRotatef $zrot 0.0 0.0 1.0

	glClear GL_COLOR_BUFFER_BIT
	glClear GL_DEPTH_BUFFER_BIT

	glBegin GL_LINES
	glVertex3d 0.0 0.0 0.0
	glVertex3d 1.0 1.0 2.0
	glEnd

	glMapGrid2d 20 0 1 20 0 1 
	glEvalMesh2 GL_FILL 0 20 0 20

	$win swapBuffer
}


bind .gl <Button-1> { set x1 %x ; set y1 %y }
bind .gl <B1-Motion> { 
	set yrot1 [expr $yrot1 + [expr %x - $x1] ]
	set xrot1 [expr $xrot1 + [expr %y - $y1] ]
	set x1 %x
	set y1 %y
	.gl redraw
}

bind .gl1 <Button-1> { set x2 %x ; set y2 %y }
bind .gl1 <B1-Motion> { 
	set yrot2 [expr $yrot2 + [expr %x - $x2] ]
	set xrot2 [expr $xrot2 + [expr %y - $y2] ]
	set x2 %x
	set y2 %y
	.gl1 redraw
}

set animateState 0
set animSpeed 20
toplevel .t
button .t.b -command startAnimate -text start
button .t.bst -command stopAnimate -text stop
pack .t.b .t.bst -side top

scale .t.s -from 1 -to 100 -variable animSpeed
scale .t.red -from 0 -to 1.0 -resolution 0.01 -variable red -command colour
scale .t.green -from 0 -to 1.0 -resolution 0.01 -variable green -command colour
scale .t.blue -from 0 -to 1.0 -resolution 0.01 -variable blue -command colour
scale .t.shine -from 0 -to 100 -variable shiny -command shinyProc

pack .t.s .t.red .t.green .t.blue .t.shine -side left

proc animate {} {
	global animateState xrot1 yrot1 xrot2 yrot2 animSpeed

	if {$animateState == 1} {
		set xrot1 [expr $xrot1 + 3.0]
		set yrot1 [expr $yrot1 + 3.0]
		set xrot2 [expr $xrot2 - 3.0]
		set yrot2 [expr $yrot2 - 3.0]
		.gl redraw
		.gl1 redraw
		after $animSpeed animate
	}
}

proc startAnimate {} {
	global animateState
	set animateState 1
	animate
}

proc stopAnimate {} {
	global animateState
	set animateState 0
}

proc speed {val} {
	global animSpeed
	set animSpeed $val
}

proc shinyProc {val} {
	.gl redraw
	.gl1 redraw
}
