package require Tkogl
package require OpenGL
package require OpenGLU


set debug [open "c:\\debug.txt" w]

proc compute_picture {} {
	global nedge nvtx edge vtx picture yrot xrot

	glMatrixMode GL_PROJECTION
	glLoadIdentity
	glOrtho -4 4 -4 4 -4 4
	glMatrixMode GL_MODELVIEW
	glLoadIdentity

	glNewList $picture GL_COMPILE
	
	glClear GL_COLOR_BUFFER_BIT
	glClear GL_DEPTH_BUFFER_BIT
	glShadeModel GL_FLAT
	glColor3f 1 1 0
	
	glBegin GL_LINES
		for {set i 0} {$i < $nedge} {incr i} {
			glVertex3d [lindex $vtx($edge($i,0)) 0] [lindex $vtx($edge($i,0)) 1] [lindex $vtx($edge($i,0)) 2]
			glVertex3d [lindex $vtx($edge($i,1)) 0] [lindex $vtx($edge($i,1)) 1] [lindex $vtx($edge($i,1)) 2]
		}
	glEnd

	glInitNames
	glPushName -1
	glPointSize 5
	glColor3f 0 1 1

	for {set i 0} {$i < $nvtx} {incr i} {
		glLoadName $i
		glBegin GL_POINTS
			glVertex3d [lindex $vtx($i) 0] [lindex $vtx($i) 1] [lindex $vtx($i) 2]
		glEnd
	}

	glEndList	
}
    
proc redraw {} {
	global xrot yrot nedge nvtx edge vtx picture

	glMatrixMode GL_PROJECTION
	glLoadIdentity
	glOrtho -4 4 -4 4 -4 4
	glMatrixMode GL_MODELVIEW
	glLoadIdentity
	glRotated $yrot 1 0 0
	glRotated $xrot 0 1 0

	glCallList $picture

	.gl swapBuffer
}

proc pick_point { x y } {
	global picture vtx pickvtx pickz debug

	set Buffer [array_int 100]
	set viewport [array_int 4]

	.gl makeCurrent

	glSelectBuffer 100 $Buffer
	glGetIntegerv GL_VIEWPORT $viewport
	
	glMatrixMode GL_PROJECTION
	glPushMatrix
	glLoadIdentity

	glRenderMode GL_SELECT	
	
	gluPickMatrix $x [expr [get_int $viewport 3] - $y] 4 4 $viewport
	glOrtho -4 4 -4 4 -4 4
	glMatrixMode GL_MODELVIEW
	glCallList $picture
	glMatrixMode GL_PROJECTION
	glPopMatrix
	glMatrixMode GL_MODELVIEW

	glFlush

	set pickvtx {}
	set pickz {}

	set nHits [glRenderMode GL_RENDER]
	puts $debug "pick_point $x $y nHits = $nHits"

	set msg ""
	set n 0

	if {$nHits != 0} {
		for {set i 0} {$i < $nHits} {incr i} {
			set msg "$msg [get_int $Buffer $n]"
			set num [get_int $Buffer $n]
			incr n
			set msg "$msg [get_int $Buffer $n]"
			incr n
			set msg "$msg [get_int $Buffer $n]"
			incr n
			for {set j 0} {$j < $num} {incr j} {
				set msg "$msg [get_int $Buffer $n]"
				incr n 
			}
		}

		puts $debug $msg
	}

	if {$nHits != 0} { 
		set pickvtx [get_int $Buffer 6]
		set pickz [expr double([get_int $Buffer 3])]
		puts $debug "  $pickvtx $pickz"
	}
}

proc move_point { x y } {
    global pickvtx pickz vtx
    if {$pickvtx=={}} return
    set vtx($pickvtx) [.gl unproject $x $y $pickz]
    compute_picture
    redraw
}


pack [OGLwin .gl -refresh redraw] -expand yes -fill both

set nvtx 0
foreach x {-1 1} {
    foreach y {-1 1} {
	foreach z {-1 1} {
	    set vtx($nvtx) "$x $y $z"
	    incr nvtx
	}
    }
}

set nedge 0
for {set i 0} {$i < 8} { incr i} {
    for {set j [expr $i+1]} {$j < 8} {incr j} {
	set count 0 
	for {set k 0} {$k < 3} {incr k} {
	    if [lindex $vtx($i) $k]==[lindex $vtx($j) $k] { incr count } 
	}
	if $count==2 {
	    set edge($nedge,0) $i 
	    set edge($nedge,1) $j
	    incr nedge
	}
    }
}

set picture [glGenLists 1]

compute_picture
	
set xrot 10 
set yrot 10

bind .gl <Button-1> { set x %x ; set y %y }
bind .gl <B1-Motion> { 
    incr xrot [expr %x - $x] 
    incr yrot [expr %y - $y] 
    set x %x
    set y %y
    redraw
}
bind .gl <Button-3> { pick_point %x %y }
bind .gl <B3-Motion> { move_point %x %y }
redraw


