#ifndef _TK_OGL
#define _TK_OGL

#include <tk.h>
#if defined(__WIN32__) || defined(_WIN32)
#   define WIN32_LEAN_AND_MEAN
#   include <windows.h>
#   undef WIN32_LEAN_AND_MEAN
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#include <assert.h>
#if defined(__WIN32__) || defined(_WIN32)
#   include "tkWinInt.h"
#   if defined(_MSC_VER)
#	    define EXPORT(a,b) __declspec(dllexport) a b
#	    define DllEntryPoint DllMain
#   else
#	    if defined(__BORLANDC__)
#               include <float.h>
#	        define EXPORT(a,b) a _export b
#	    else
#	        define EXPORT(a,b) a b
#	    endif
#   endif
#else
#   define EXPORT(a,b) a b
#   include <GL/glx.h>
#   include <X11/Xatom.h>		/* for XA_RGB_DEFAULT_MAP atom */
#   include <X11/Xmu/StdCmap.h>	/* for XmuLookupStandardColormap() */
#endif

/*
 * A data structure of the following type is kept for each glxwin
 * widget managed by this file:
 */

typedef struct {
    Tk_Window tkwin;		/* Window that embodies the glxwin.  NULL
   		                     * means window has been deleted but
				 * widget record hasn't been cleaned
                                         * up yet. */
    Display *display;		/* X's token for the window's display. */
    Tcl_Interp *interp;		/* Interpreter associated with widget. */

    int width;			/* Width to request for window.  <= 0 means
				 * don't request any size. */
    int height;			/* Height to request for window.  <= 0 means
			           * don't request any size. */

    int absx, absy;                     /* Absolute x and y of window */

    char* context;	                    /* Name of gl window this window will share
                                         * display lists with */

    int flags;			/* Various flags;  see below for
				 * definitions. */

    int doubleBuffer,           /* Various options that control visual alloc. */
        depthSize, 
        stencilSize, 
        alphaSize,
        accumSize;

    int updatePending;		/* Set to 1 when redraw needed */

    int redrawList;		/* Number of redraw display list (or -1) */

    double aspectRatio; 	/* If not 0, represents a fraction width/height
				   which should be maintained when setting
				   the viewport during the event loop */

    char * refresh;
	
	Tk_3DBorder bgBorder;	/* Used for drawing background. */

#if defined(__WIN32__) || defined(_WIN32)

    /*
     * Information required to manage the OpenGl window in WIN32 environment
     */

    HPALETTE hPalette;
    HWND hwnd;
    HDC hdc;
    HGLRC hrc;
    TkWinDCState state;
#else
    GLXContext cx;              /* The GL X context */
#endif

} OGLwin;


typedef int (TkOGLExtProc) (Tcl_Interp* interp, int argc, char** argv);

EXPORT (int,RegisterTkOGLExtension) (Tcl_Interp* interp, char* extname, TkOGLExtProc* extproc);

#endif /* _TK_OGL */
